#define _XOPEN_SOURCE /* for strptime */
#define _BSD_SOURCE /* for snprintf */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <pwd.h>
#include <dirent.h>
#include "getopt.h"
#include "texinfo_parser.h"

#define BUFFER_SIZE 1024
#define DEFAULT_ENTRY_COUNT 14
#define JOURNAL_DIR "journal"
#define OPTIONS "c:f:hr:s:t:v"

/* The list of long options */
static struct option long_options[] =
{
    { "count", required_argument, NULL, 'c' },
    { "from", required_argument, NULL, 'f' },
    { "help", no_argument, NULL, 'h' },
    { "root", required_argument, NULL, 'r' },
    { "skip", required_argument, NULL, 's' },
    { "to", required_argument, NULL, 't' },
    { "version", no_argument, NULL, 'v' },
    { NULL, no_argument, NULL, '\0' }
};


/* A useful struct for sorting strings of numbers */
struct num_string
{
    int value;
    char string[1];
};

/* An association list */
struct alist
{
    struct alist *next;
    char *name;
    char *value;
};


/* Table for converting nibbles from from hexadecimal to decimal */
static const char nibble_values[] =
{
    /* 0x30 */  0,  1,  2,  3,   4,  5,  6,  7,
    /* 0x38 */  8,  9, -1, -1,  -1, -1, -1, -1,
    /* 0x40 */ -1, 10, 11, 12,  13, 14, 15, -1,
    /* 0x48 */ -1, -1, -1, -1,  -1, -1, -1, -1,
    /* 0x50 */ -1, -1, -1, -1,  -1, -1, -1, -1,
    /* 0x58 */ -1, -1, -1, -1,  -1, -1, -1, -1,
    /* 0x60 */ -1, 10, 11, 12,  13, 14, 15, -1
};

/* Table of valid characters in a CGI query string */
static const char query_chars[] =
{
    /* 0x20 */ 0, 1, 0, 0,  1, 0, 1, 1,   1, 1, 1, 1,  1, 1, 1, 1,
    /* 0x30 */ 1, 1, 1, 1,  1, 1, 1, 1,   1, 1, 1, 1,  0, 1, 0, 1,
    /* 0x40 */ 1, 1, 1, 1,  1, 1, 1, 1,   1, 1, 1, 1,  1, 1, 1, 1,
    /* 0x50 */ 1, 1, 1, 1,  1, 1, 1, 1,   1, 1, 1, 0,  0, 0, 0, 1,
    /* 0x60 */ 0, 1, 1, 1,  1, 1, 1, 1,   1, 1, 1, 1,  1, 1, 1, 1,
    /* 0x70 */ 1, 1, 1, 1,  1, 1, 1, 1,   1, 1, 1, 0,  0, 0, 1, 0
};

/* Date formats suitable for strptime's use */
static const char *const date_formats[] =
{
    "%D",
    "%F",
    NULL
};

/* A useful struct for sorting strings of numbers */
typedef struct num_string *num_string_t;

/* The year entries */
static num_string_t *years = NULL;
static int year_count = 0;

/* The day entries */
static num_string_t *days = NULL;
static int day_count = 0;

/* The texinfo parser */
static texinfo_parser_t parser;


static int parsedate(const char *string, struct tm *tm_out)
{
    struct timeval now;
    struct tm *tm;
    char *end;
    time_t when;
    int i;

    /* Get the current time */
    if (gettimeofday(&now, NULL) < 0) {
	perror("gettimeofday failed");
	exit(1);
    }

    /* Convert it to local time */
    when = (time_t)now.tv_sec;
    tm = localtime(&when);
    if (tm == NULL) {
	perror("localtime failed");
	exit(1);
    }

    for (i = 0; date_formats[i]; i++) {
	memcpy(tm_out, tm, sizeof(struct tm));
	end = strptime(string, date_formats[i], tm_out);
	if (end && *end == '\0') {
	    return 0;
	}
    }

    return -1;
}

static inline int nibbletodec(int ch)
{
    return (ch >= 0x30 && ch < 0x68) ? nibble_values[ch - 0x30] : -1;
}

static inline int isquery(int ch)
{
    return (ch >= 0x20 && ch < 0x7f) ? query_chars[ch - 0x20] : 0;
}

/* Parses a cgi-bin query string into names and values.  A pointer to
 * the mangled string data is available as the name of the first entry
 * in the list. */
static struct alist *
parsequery(const char *query)
{
    char *string, *out, *name, *value;
    int ch, esc_state, digit, num;
    struct alist *alist, *last, *entry;
    size_t len;

    /* Allocate room for the parsed results */
    len = strlen(query);
    string = malloc(len + 1);

    /* Split the query into names and values */
    name = out = string;
    value = NULL;
    alist = last = NULL;
    esc_state = 0;
    do {
        ch = *query++;
        switch (esc_state) {
        case 0:
            if (ch == '%') {
                esc_state = 1;
                continue;
            }

            if (value == NULL && ch == '=') {
                /* Make sure the name contains at least one
                 * character. */
                if (name == out) {
                    goto err;
                }

                *out++ = '\0';
                value = out;
                continue;
            }

            if (ch == '&' || ch == '\0') {
                /* Make sure the name or value contains at least one
                 * character. */
                if (name == out || value == out) {
                    goto err;
                }
                *out++ = '\0';

                /* Allocate another alist entry */
                if ((entry = malloc(sizeof(struct alist))) == NULL) {
                    goto err;
                }
                entry->next = NULL;
                entry->name = name;
                entry->value = value;

                /* Add it to the end of the list */
                if (last) {
                    last->next = entry;
                } else {
                    alist = entry;
                }
                last = entry;

                /* Prepare for the next entry */
                name = out;
                value = NULL;
                continue;
            }

            if (!isquery(ch)) {
                goto err;
            }
            break;

        case 1:
            if ((digit = nibbletodec(ch)) < 0) {
                goto err;
            }
            num = digit;
            esc_state++;
            continue;

        case 2:
            if ((digit = nibbletodec(ch)) < 0) {
                goto err;
            }

            if ((ch = (num << 4) | digit) == 0) {
                goto err;
            }

            esc_state = 0;
            break;
        }

        *out++ = ch;
    } while (ch != 0);

    /* Don't allow partial escaped characters */
    if (esc_state != 0) {
        goto err;
    }

    return alist;

  err:
    /* Discard the entries */
    while ((entry = alist) != NULL) {
        alist = entry->next;
        free(entry);
    }

    /* And free the mangled string */
    free(string);
    return NULL;
}

/* Allocates and initializes a new number string */
static num_string_t num_string_alloc(char *string)
{
    num_string_t self;
    char *pointer;
    int ch;

    /* Make sure that 'num' consists entirely of digits */
    for (pointer = string; (ch = *pointer) != 0; pointer++) {
	if (!isdigit(ch)) {
	    return NULL;
	}
    }

    /* Allocate enough space for the num_string and its value */
    if ((self = (num_string_t)malloc(sizeof(struct num_string) + pointer - string)) == NULL) {
	return NULL;
    }

    /* Fill in the blanks */
    self -> value = atoi(string);
    memcpy(self -> string, string, pointer - string + 1);
    return self;
}

/* Frees a num_string */
static void num_string_free(num_string_t self)
{
    free(self);
}

/* Compares two number strings for ascending qsort order */
static int num_string_cmp(const void *num1, const void *num2)
{
    num_string_t self = *(num_string_t *)num1;
    num_string_t num_string = *(num_string_t *)num2;

    if (self -> value < num_string -> value) {
	return -1;
    }

    if (self -> value > num_string -> value) {
	return 1;
    }

    return 0;
}
/* Compares two number strings for descending qsort order */
static int num_string_rev_cmp(const void *num1, const void *num2)
{
    num_string_t self = *(num_string_t *)num1;
    num_string_t num_string = *(num_string_t *)num2;

    if (self -> value < num_string -> value) {
	return 1;
    }

    if (self -> value > num_string -> value) {
	return -1;
    }

    return 0;
}


/* Append the contents of 'in' to the end of 'out' */
static int concat(FILE *in, FILE *out)
{
    char buffer[BUFFER_SIZE];
    size_t size;

    /* Keep copying until we run out of stuff */
    for (;;) {
	/* Read some stuff */
	if ((size = fread(buffer, 1, BUFFER_SIZE, in)) == 0) {
	    /* FIX THIS: should check return codes */
	    return 0;
	}

	/* Write it */
	fwrite(buffer, 1, size, out);
	/* FIX THIS: should check return codes */
    }
}

/* Prints a nice HTML version of an error message */
static void print_error(char *message, FILE *out)
{
    fprintf(out,
	    "    <HR>\n"
	    "\n"
	    "    <TABLE width=100%% cellspacing=0 cellpadding=3>\n"
	    "      <TR>\n"
	    "        <TD>%s</TD>\n"
	    "      </TR>\n"
	    "    </TABLE>\n",
	    message);
}

/* This is called when a texinfo file has been successfully parsed */
static void parser_cb(void *rock, char *string)
{
    FILE *out = (FILE *)rock;
    fwrite(string, 1, strlen(string), out);
}

/* Convert the texinfo file 'in' (named 'filename') to HTML and append
 * it to the end of 'out' */
static int transform(char *filename, FILE *in, FILE *out)
{
    char buffer[BUFFER_SIZE];
    size_t length;

    /* Reset the parser */
    texinfo_parser_reset(parser, filename, parser_cb, out);

    /* Keep copying until we run out of stuff */
    for (;;) {
	/* Read some stuff */
	length = fread(buffer, 1, BUFFER_SIZE, in);

	/* Run it through the parser */
	if (texinfo_parser_parse(parser, buffer, length) < 0) {
	    abort();
	}

	/* Watch for EOF */
	if (length == 0) {
	    return 0;
	}
    }
}


/* Prints a nice HTML version of an I/O error message */
static void print_errno(char *filename, int error_num, FILE *out)
{
    char buffer[BUFFER_SIZE];

    sprintf(buffer, "%s: unable to open: %s", filename, strerror(error_num));
    print_error(buffer, out);
}

/* Copies a single entry from the filename to out */
static int add_entry(num_string_t year, num_string_t day,
		     struct tm *tm, FILE *out)
{
    char buffer[BUFFER_SIZE];
    FILE *file;

    /* Construct a filename from the year and day */
    snprintf(buffer, sizeof(buffer), "%s/%s", year->string, day->string);

    /* If we can't open the file then skip it */
    if ((file = fopen(buffer, "r")) == NULL) {
	print_errno(buffer, errno, out);
	return -1;
    }

    /* Print out a header for the entry */
    strftime(buffer, sizeof(buffer), "%A, %e %b %Y", tm);
    fprintf(out,
	    "    <HR>\n"
	    "\n"
	    "    <TABLE width=100%% cellspacing=0 cellpadding=3>\n"
	    "      <TR>\n"
	    "        <TH>%s</TH>\n"
	    "      </TR>\n"
	    "      <TR>\n"
	    "        <TD>\n", buffer);
    transform(day->string, file, out);
    fprintf(out,
	    "        </TD>\n"
	    "      </TR>\n"
	    "    </TABLE>\n");
    fclose(file);
    return 0;
}

/* Adds up to 'count' entries from the year, returning the number of entries */
static int add_year_entries(num_string_t year, int skip, int count,
			    int ascending, struct tm *first, struct tm *last,
			    FILE *out, struct tm *last_out)
{
    DIR *dir;
    struct dirent *entry;
    int index, result;
    int sum = 0;
    int num = 0;
    num_string_t num_string;
    struct tm when;

    /* Try to open the year as a directory */
    if ((dir = opendir(year -> string)) == NULL) {
	print_errno(year -> string, errno, out);
	return 0;
    }

    /* Go through the entries and look for day entries */
    while ((entry = readdir(dir)) != NULL) {
	/* Try to make a num_string entry for this day */
	if ((num_string = num_string_alloc(entry -> d_name)) != NULL) {
	    days = realloc(days, sizeof(num_string_t) * (day_count + 1));
	    days[day_count++] = num_string;
	}
    }

    /* We're done with the year now */
    closedir(dir);

    /* Sort the days */
    qsort(days, day_count, sizeof(num_string_t),
	  ascending ? num_string_cmp : num_string_rev_cmp);

    /* Go through the days and add their entries */
    for (index = 0; index < day_count && (!count || num < count); index++) {
	result = 0;

	/* Convert the entry's date into broken-down time */
	memset(&when, 0, sizeof(when));
	when.tm_year = year->value - 1900;
	when.tm_mday = days[index]->value;
	(void)mktime(&when);

	/* Skip it if it's outside the range of dates */
	if (first) {
	    if (ascending) {
		if (when.tm_mon < first->tm_mon ||
		    (when.tm_mon == first->tm_mon &&
		     when.tm_mday < first->tm_mday)) {
		    continue;
		}
	    } else {
		if (when.tm_mon > first->tm_mon ||
		    (when.tm_mon == first->tm_mon &&
		     when.tm_mday > first->tm_mday)) {
		    continue;
		}
	    }
	}

	/* Are we done? */
	if (last) {
	    if (ascending) {
		if (when.tm_year > last->tm_year ||
		    (when.tm_year == last->tm_year &&
		     (when.tm_mon > last->tm_mon ||
		      (when.tm_mon == last->tm_mon &&
		       when.tm_mday > last->tm_mday)))) {
		    break;
		}
	    } else {
		if (when.tm_year < last->tm_year ||
		    (when.tm_year == last->tm_year &&
		     (when.tm_mon < last->tm_mon ||
		      (when.tm_mon == last->tm_mon &&
		       when.tm_mday < last->tm_mday)))) {
		    break;
		}
	    }
	}

	/* Add the entry if we're done skipping */
	if (skip < 1) {
	    result = add_entry(year, days[index], &when, out);
	}

	num_string_free(days[index]);
	first = NULL;

	if (result >= 0) {
	    sum++;
	    if (skip > 0) {
		skip--;
	    } else {
		num++;
	    }
	}
    }

    /* Clean up */
    free(days);
    days = NULL;
    day_count = 0;
    memcpy(last_out, &when, sizeof(when));
    return sum;
}

/* Copies a range of entries to the given file */
static int add_entries(int skip, int count, int ascending,
		       struct tm *first, struct tm *last, FILE *out,
		       struct tm *last_out)
{
    DIR *dir;
    struct dirent *entry;
    num_string_t num_string;
    int index, num;
    int sum = 0;

    /* Read the current directory */
    if ((dir = opendir(".")) == NULL) {
	print_errno(".", errno, out);
	return 0;
    }

    /* Go through the entries and look for year directories */
    while ((entry = readdir(dir)) != NULL) {
	/* Try to make a num_string entry for this one */
	if ((num_string = num_string_alloc(entry -> d_name)) != NULL) {
	    years = realloc(years, sizeof(num_string_t) * (year_count + 1));
	    years[year_count++] = num_string;
	}
    }

    /* We're done with the dir now */
    closedir(dir);

    /* Sort the years as appropriate */
    qsort(years, year_count, sizeof(num_string_t),
	  ascending ? num_string_cmp : num_string_rev_cmp);

    /* Go through the years and add entries as appropriate */
    for (index = 0; index < year_count && (!count || sum < count); index++) {
	if (first) {
	    if (ascending) {
		if (years[index]->value < first->tm_year + 1900) {
		    continue;
		}
	    } else { 
		if (years[index]->value > first->tm_year + 1900) {
		    continue;
		}
	    }
	}

	/* Are we done? */
	if (last) {
	    if (ascending) {
		if (years[index]->value > last->tm_year + 1900) {
		    break;
		}
	    } else {
		if (years[index]->value < last->tm_year + 1900) {
		    break;
		}
	    }
	}

	/* Print out some number of entries from the given year */
	num = add_year_entries(years[index], skip, count - sum,
			       ascending, first, last, out, last_out);
	num_string_free(years[index]);

	/* We've found the first entry, so we can stop checking for it. */
	first = NULL;

	if (num < skip) {
	    skip -= num;
	} else {
	    sum += num - skip;
	    skip = 0;
	}
    }

    return 0;
}

/* Prints a usage message */
static void usage(int argc, char *argv[])
{
    fprintf(
	stderr,
 	"usage: %s [OPTION]...\n"
	"  -r dir,  --root=dir\n"
	"  -c num,  --count=num\n"
	"  -f date, --from=date\n"
	"  -s num,  --skip=num\n"
	"  -t date, --to=date\n"
	"  -v,      --version\n"
	"  -h,      --help\n", argv[0]);
}


/* Construct a web page from journal entries */
int main(int argc, char *argv[])
{
    int choice;
    int count = 0;
    struct tm from, to, final, *first, *last;
    char *pointer;
    int skip = 0, have_from = 0, have_to = 0, ascending;
    struct alist *alist, *ap;
    struct passwd *user;
    char buffer[256];
    FILE *file;

    /* Create a texinfo parser for future use */
    parser = texinfo_parser_alloc();

    /* If additional arguments are on the command-line then parse
     * them. */
    if (argc > 1) {
        /* Parse the command-line */
        while ((choice = getopt_long(argc, argv, OPTIONS, long_options, NULL)) != -1) {
            switch (choice) {
            case 'c':
                /* --count or -c */
                count = strtol(optarg, &pointer, 0);
                if (count < 1 || *pointer != '\0') {
                    fprintf(stderr, "Error: invalid argument to count: %s\n", optarg);
                    usage(argc, argv);
                    exit(1);
                }

                break;

            case 'f':
                /* --from or -f */
                if (parsedate(optarg, &from) < 0) {
                    fprintf(stderr, "Error: invalid time string: %s\n", optarg);
                    usage(argc, argv);
                    exit(1);
                }
                have_from = 1;
                break;

            case 'r':
                /* --root or -r */
                /* Try to change to that directory */
                if (chdir(optarg) < 0) {
                    perror("chdir(): failed");
                    exit(1);
                }

                break;

            case 's':
                /* --skip or -s */
                skip = strtol(optarg, &pointer, 0);
                if (*pointer != '\0') {
                    fprintf(stderr, "Error: invalid argument to skip: %s\n", optarg);
                    usage(argc, argv);
                    exit(1);
                }

                break;

            case 't':
                /* --to or -t */
                if (parsedate(optarg, &to) < 0) {
                    fprintf(stderr, "Error: invalid time string: %s\n", optarg);
                    usage(argc, argv);
                    exit(1);
                }
                have_to = 1;
                break;

            case 'v':
                /* --version or -v */
                printf("%s version %s\n", PACKAGE, VERSION);
                exit(0);

            case 'h':
                /* --help or -h */
                usage(argc, argv);
                exit(0);

            case '?':
                usage(argc, argv);
                exit(1);

            default:
                abort();
            }
        }
    } else if ((pointer = getenv("QUERY_STRING")) != NULL) {
        /* We're running in a cgi-bin environment. */
	if (*pointer == '\0') {
	    printf("Content-type: text/plain\n\n"
		   "error: no user specified\n");
	    exit(0);
	}

	/* Parse the query. */
        user = NULL;
        if ((alist = parsequery(getenv("QUERY_STRING"))) == NULL) {
            printf("Content-type: text/plain\n\n"
                   "error: invalid query string: %s\n",
                   getenv("QUERY_STRING"));
            exit(0);
        }

        for (ap = alist; ap; ap = ap->next) {
            if (strcmp(ap->name, "count") == 0) {
                if (ap->value == NULL) {
                    printf("Content-type: text/plain\n\n"
                           "error: option 'count' requires a value\n");
                    exit(0);
                }

                count = strtol(ap->value, &pointer, 0);
                if (count < 1 || *pointer != '\0') {
                    printf("Content-type: text/plain\n\n"
                           "error: invalid count: %s\n", ap->value);
                    exit(0);
                }
            } else if (strcmp(ap->name, "from") == 0) {
                if (ap->value == NULL) {
                    printf("Content-type: text/plain\n\n"
                           "error: option 'from' requires a value\n");
                    exit(0);
                }

                if (parsedate(ap->value, &from) < 0) {
                    printf("Content-type: text/plain\n\n"
                           "error: invalid time string: %s\n", ap->value);
                    exit(0);
                }

		have_from = 1;
            } else if (strcmp(ap->name, "skip") == 0) {
                if (ap->value == NULL) {
                    printf("Content-type: text/plain\n\n"
                           "error: option 'skip' requires a value\n");
                    exit(0);
                }

                skip = strtol(ap->value, &pointer, 0);
                if (*pointer != '\0') {
                    printf("Content-type: text/plain\n\n"
                           "error: invalid argument to skip: %s\n", ap->value);
                    exit(0);
                }
            } else if (strcmp(ap->name, "to") == 0) {
                if (ap->value == NULL) {
                    printf("Content-type: text/plain\n\n"
                           "error: option 'to' requires a value\n");
                    exit(0);
                }

                if (parsedate(ap->value, &to) < 0) {
                    printf("Content-type: text/plain\n\n"
                           "error: invalid time string: %s\n", ap->value);
                    exit(0);
                }

		have_to = 1;
            } else if (strcmp(ap->name, "user") == 0) {
                if (ap->value == NULL) {
                    printf("Content-type: text/plain\n\n"
                           "error: option 'user' requires a value\n");
                    exit(0);
                }

                /* Look up the user */
                if ((user = getpwnam(ap->value)) != NULL) {
                    snprintf(buffer, sizeof(buffer), "%s/" JOURNAL_DIR, user->pw_dir);
		    if (chdir(buffer) < 0) {
                        user = NULL;
                    }
                }

                /* Report the same error for all user-related problems
                 * so that we don't give away any potentially useful
                 * information. */
                if (!user) {
                    printf("Content-type: text/plain\n\n"
                           "error: no journal available for user %s\n", ap->value);
                    exit(0);
                }
            } else {
                printf("Content-type: text/plain\n\n"
                       "error: unsupport query option: %s\n", ap->name);
                exit(0);
            }
        }

        /* Bail if no user given */
        if (!user) {
            printf("Content-type: text/plain\n\n"
                   "error: no user specified\n");
            exit(0);
        }

        /* Clean up */
        free(alist->name);
        while ((ap = alist) != NULL) {
            alist = ap->next;
            free(ap);
        }
    }

    /* Copy the header to stdout */
    if ((file = fopen("header.html", "r")) == NULL) {
        printf("Content-type: text/plain\n\n"
               "error: unable to open header.html: %s\n", strerror(errno));
        exit(0);
    }

    /* Print the content-type information */
    printf("Content-type: text/html\n\n");

    concat(file, stdout);
    fclose(file);

    /* Determine whether to report the results in ascending or
     * descending order. */
    ascending = 0;
    if (have_from) {
        if (have_to) {
	    first = &from;
	    last = &to;

	    /* Determine whether to list dates from oldest to newest
	     * or the other way around. */
            if (from.tm_year < to.tm_year ||
                (from.tm_year == to.tm_year &&
                 (from.tm_mon < to.tm_mon ||
                  (from.tm_mon == to.tm_mon &&
                   from.tm_mday < to.tm_mday)))) {
                ascending = 1;
            }
        } else {
            print_error("Error: listing entries from a date is not yet supported", stdout);
            goto out;
        }
    } else {
	/* If no count given then use the default. */
	count = count ? count : DEFAULT_ENTRY_COUNT;
        first = have_to ? &to : NULL;
        last = NULL;
    }

    /* Copy the entries into the middle */
    memset(&final, 0, sizeof(final));
    add_entries(skip, count, ascending, first, last, stdout, &final);

    /* Add a link to older items */
    if (user && final.tm_mday && (pointer = getenv("SCRIPT_NAME")) != NULL) {
	final.tm_mday--;
	if (mktime(&final) != (time_t)-1) {
	    strftime(buffer, sizeof(buffer), "%F", &final);
	    printf("    <HR>\n"
		   "\n"
		   "    <TABLE width=100%% cellspacing=0 cellpadding=3>\n"
		   "      <TR>\n"
		   "        <TH><A HREF=\"%s?user=%s&to=%s\">"
		   "older entries</A></TH>\n"
		   "      </TR>\n"
		   "    </TABLE>\n",
		   pointer, user->pw_name, buffer);
	}
    }

out:
    /* Copy the footer to stdout */
    if ((file = fopen("footer.html", "r")) == NULL) {
        print_errno("footer.html", errno, stdout);
        return 0;
    }

    concat(file, stdout);
    fclose(file);
    return 0;
}
